using System;
using System.Collections;
using System.Drawing;
using CenterSpace.NMath.Core;
using CenterSpace.NMath.Matrix;
using System.Windows.Forms;

namespace SpectralNet
{
	/// <summary>
	/// This class is the C# representation of the chemical interaction graph.
	/// It contains vertices, edges, and a collection of SubGraphs that are
	/// the connected components of the graph.
	/// NOTE that addVertex/addEdge does NOT update the adj matrices of the subgraphs
	///      since it would be so expensive.  This is done in .UpdateComponents()
	/// To calculate eigenspectra, first run .UpdateComponents() then run .CalculateVertexData() on a component.
	/// </summary>
	public class ChemGraph 
	{
		/// <summary>
		/// The vertex array that goes along with this adjacency matrix
		/// Access .Component to find a vertex's subgraph, but only if you've run .UpdateComponents()!
		/// </summary>
		Hashtable _vertices;

		public ProgressBar pb;

		public bool submittedViaCommandLine;

		/// <summary>
		/// The list of edges in this graph.
		/// </summary>
		ArrayList _edges;

		/// <summary>
		///  A collection of ConnectedComponent objects that comprise this Graph.
		/// </summary>
		ArrayList _components;

		/// <summary>
		/// List of files currently being used, for housekeeping.
		/// </summary>
		ArrayList _files;

		/// <summary>
		/// Whether or not this graph is bipartite.  Unused?
		/// </summary>
		public bool isBipartite;

		/// <summary>
		/// Whether or not this graph is unweighted.
		/// </summary>
		public bool isUnweighted;

		/// <summary>
		/// The current display type of this graph, normal or laplacianLLE.
		/// </summary>
		displayType _display;

		/// <summary>
		/// Indicates whether the GUI should display edges connected to the current vertex
		/// </summary>
		public bool drawConnectedEdges;

		/// <summary>
		/// Indicates whether the GUI should display edges not connected to the current vertex
		/// </summary>
		public bool drawOtherEdges;

		/// <summary>
		/// Holds options like whether the user wants us to calculate Fruchterman-Reingold or path lengths, etc.
		/// </summary>
		public UserOptions options;

		/// <summary>
		/// The width of the main display image
		/// </summary>
		public int mainDisplayWidth;

		/// <summary>
		/// The height of the main display image
		/// </summary>
		public int mainDisplayHeight;

		/// <summary>
		/// The width of the small images (i.e., for degree distribution, eigenvalues, eigenvectors, etc.)
		/// </summary>
		public int smallDisplayWidth;

		/// <summary>
		/// The height of the small images (i.e., for degree distribution, eigenvalues, eigenvectors, etc.)
		/// </summary>
		public int smallDisplayHeight;

		/// <summary>
		/// The URL location of the main display image (Fruchterman-Reingold)
		/// </summary>
		public string mainDisplayLocation;

		/// <summary>
		/// The URL location of the PCA display image
		/// </summary>
		public string pcaDisplayLocation;

		/// <summary>
		/// The currently selected vertex
		/// </summary>
		private Vertex _selectedVertex;
		
		/// <summary>
		/// Whether PCA has been performed on the input data (always false for generated graphs)
		/// </summary>
		public bool pcaCalculated;

		/// <summary>
		/// The scores matrix if PCA has been performed on the input data
		/// </summary>
		public DoubleMatrix pcaScores;

		/// <summary>
		/// The minimum edge weight of all edges in this graph
		/// </summary>
		public double minEdgeWeight
		{
			get
			{
				double result = double.MaxValue;
				
				foreach(Edge e in this.Edges)
				{
					if (e.weight < result) result = e.weight;
				}

				return result;
			}
		}

		/// <summary>
		/// The maximum edge weight of all edges in this graph
		/// </summary>
		public double maxEdgeWeight
		{
			get
			{
				double result = double.MinValue;
				
				foreach(Edge e in this.Edges)
				{
					if (e.weight > result) result = e.weight;
				}

				return result;
			}
		}

		/// <summary>
		/// Property wrapper for _display.
		/// </summary>
		public Vertex SelectedVertex
		{
			get { return _selectedVertex; }
			set { _selectedVertex = value; }
		}

		/// <summary>
		/// Property wrapper for _display.
		/// </summary>
		public displayType DisplayType
		{
			get { return _display; }
			set { _display = value; }
		}

		/// <summary>
		/// Returns the number of vertices in this graph.
		/// </summary>
		public int NumVertices 
		{
			get { return _vertices.Count; }
		}

		/// <summary>
		/// Returns the number of edges in this graph.
		/// </summary>
		public int NumEdges 
		{
			get { return _edges.Count; }
		}

		/// <summary>
		/// Property wrapper for _vertices.
		/// </summary>
		public Hashtable Vertices 
		{
			get { return _vertices; }
		}

		/// <summary>
		/// Property wrapper for _edges.
		/// </summary>
		public ArrayList Edges 
		{
			get { return _edges; }
		}

		/// <summary>
		/// Property wrapper for _components.
		/// </summary>
		public ArrayList Components 
		{
			get { return _components; }
		}

		/// <summary>
		/// Returns the number of connected components that comprise this graph.
		/// </summary>
		public int NumComponents 
		{
			get { return _components.Count; }
		}

		/// <summary>
		/// Gets a vertex by ID.
		/// </summary>
		/// <param name="ID">The ID number of the vertex you want.</param>
		/// <returns>The vertex desired.</returns>
		public Vertex getVertex(int ID) 
		{
			return (Vertex)_vertices[ID];
		}

		/// <summary>
		/// Gets a vertex by label (slow but sometimes necessary).
		/// </summary>
		/// <param name="label">The label of the vertex you want.</param>
		/// <returns>The vertex desired.</returns>
		public Vertex getVertex(string label) 
		{
			foreach(DictionaryEntry de in _vertices)
			{
				Vertex v = (Vertex) de.Value;

				if (v.Label == label) return v;
			}

			return null;
		}

		/// <summary>
		/// Adds a file to _files for housekeeping later.
		/// </summary>
		/// <param name="s">The path of the file.</param>
		public void addFile(string s) 
		{
			_files.Add(s);
		}

		/// <summary>
		/// Constructor for a ChemGraph.
		/// </summary>
		/// <param name="size">The expected size of the graph (non-binding).</param>
		/// <param name="da">The associated web page class.</param>
		public ChemGraph(int size, int displayWidth, int displayHeight) 
		{
			_vertices = new Hashtable(size);
			_edges = new ArrayList();
			_files = new ArrayList();
			isBipartite = true;
			drawConnectedEdges = true;
			drawOtherEdges = true;
			options = new UserOptions();
			this.mainDisplayWidth = displayWidth;
			this.mainDisplayHeight = displayHeight;
			this.smallDisplayHeight = 280;
			this.smallDisplayWidth = 450;
			pcaCalculated = false;
			submittedViaCommandLine = false;
		}

		/// <summary>
		/// Updates connected component information (to be run after graph building once)
		/// NOTE: This has been changed from DFS to BFS to stop stack overflows
		/// </summary>
		public void UpdateComponents() 
		{
			_components = new ArrayList();

			foreach(DictionaryEntry de in _vertices) 
			{
				Vertex v = (Vertex)(de.Value);

				if (v.marked == false) 
				{

					//
					// BFS
					ConnectedComponent cc = new ConnectedComponent();
					
					//
					// Set up BFS starting at node curVertex
					//
					Queue q = new Queue();
					v.marked = true;
					q.Enqueue(v);

					//
					// Do the BFS
					//
					while (q.Count > 0)
					{
						Vertex u = (Vertex)q.Dequeue();
						cc.addVertex(u);

						foreach (Vertex w in u.ConnectedTo)
						{
							if (w.marked == false)
							{
								w.marked = true;
								q.Enqueue(w);
							}
						}
					}

					//
					// Done with BFS, do accounting
					//
					cc.createAdjMatrix(cc.Vertices.Count);
					cc.Graph = this;

					_components.Add(cc);
				} 
				else 
				{
					// Do nothing
				}
			}

			foreach(ConnectedComponent c in _components) 
			{
				if (c.NumVertices > 3) 
				{
					c.calculateVertexComponentInformation(this.pb);
				}
			}
		}

		/// <summary>
		/// Collapses an array of nodes into a parent node
		/// </summary>
		/// <param name="nodes">The ArrayList of nodes to collapse.</param>
		/// <returns>The new parent vertex</returns>
		public Vertex collapseNodes(string name, ArrayList nodes)
		{
			if (nodes.Count < 2)
			{
				throw new Exception("ChemGraph.collapseNodes: Cannot collapse less than two nodes into a parent node.");
			}

			// (TURNED OFF FOR NOW)
			// Don't allow sub-collapsing for now
			//
			//foreach(Vertex v in nodes)
			//{
			//	if (v.ChildVertices.Count > 0)
			//	{
			//		throw new Exception("ChemGraph.collapseNodes: Cannot sub-collapse nodes.");
			//	}
			//}

			//
			// Make sure nodes are in the same connected component
			//
			ConnectedComponent firstcc = ((Vertex)nodes[0]).ConnectedComponent;
			foreach(Vertex v in nodes)
			{
				if (v.ConnectedComponent != firstcc)
				{
					throw new Exception("ChemGraph.collapseNodes: Cannot collapse nodes that aren't in the same connected component.");
				}
			}

			//
			// Create the new vertex, add into _vertices
			//
			int maxID = 0;

			foreach(int key in _vertices.Keys)
			{
				if (key > maxID) maxID = key;
			}

			Vertex newguy = new Vertex(maxID + 1, name, false);
			newguy.ConnectedComponent = ((Vertex)nodes[0]).ConnectedComponent;
			_vertices.Add(newguy.ID, newguy);

			//
			// Remove old vertices from _vertices
			//
			foreach(Vertex v in nodes)
			{
				_vertices.Remove(v.ID);
			}

			//
			// Swap edges, making sure there are no self-edges, and storing old info into WasConnectedTo
			//
			foreach(Vertex v in nodes)
			{
				v.IsChild = true;
				v.WasConnectedTo = new ArrayList();

				//
				// Set WasConnectedTo
				//
				foreach(Edge e in v.InEdges)
				{
					if (e.v1 == v)
					{
						v.WasConnectedTo.Add(e.v2);
					}
					else
					{
						v.WasConnectedTo.Add(e.v1);
					}
				}

				//
				// Swap edges and remove self-edges
				//
				ArrayList edgesToDelete = new ArrayList();
				ArrayList theseEdges = new ArrayList(v.InEdges);
				foreach(Edge e in theseEdges)
				{
					bool didSomethingWithThisEdge = false; // diagnostic to catch bug

					//
					// If this is an edge that will lead to a self-edge when collapsed
					// CASE: IN-IN
					if (nodes.Contains(e.v1) && nodes.Contains(e.v2))
					{
						// we need to remove old vertices' refs to edge no matter what
						e.v1.removeRefToEdge(e);
						e.v2.removeRefToEdge(e);

						if (!newguy.ConnectedTo.Contains(newguy))
						{
							//
							// Add a self-edge (exchange info in edge)
							e.v1 = newguy;
							e.v2 = newguy;
							
							newguy.addRefToEdge(e);
						}
						else
						{
							// Add to weight of existing self-edge
							newguy.getEdgeWithVertex(newguy).weight += e.weight;
							edgesToDelete.Add(e);
						}

						didSomethingWithThisEdge = true;
					}

					//
					// If this is an edge that will lead to a self-edge when collapsed
					// CASE: IN-NEW
					if ((nodes.Contains(e.v1) && (e.v2 == newguy))
						|| (nodes.Contains(e.v2) && (e.v1 == newguy)))
					{
						if (!newguy.ConnectedTo.Contains(newguy))
						{
							//
							// Add a self-edge
							if (e.v1 == newguy)
							{
								e.v2.removeRefToEdge(e);
								e.v2 = newguy;
							}
							else
							{
								e.v1.removeRefToEdge(e);
								e.v1 = newguy;
							}

							// newguy already has ref to edge
						}
						else
						{
							//
							// Add to weight of existing self-edge
							if (e.v1 == newguy)
							{
								e.v2.removeRefToEdge(e);
							}
							else
							{
								e.v1.removeRefToEdge(e);
							}

							newguy.getEdgeWithVertex(newguy).weight += e.weight;
							edgesToDelete.Add(e);
						}

						didSomethingWithThisEdge = true;
					}

					//
					// CASE: NEW-NEW is impossible (we're cycling through edges from IN vertices)

					//
					// CASE: OUT-NEW is impossible (we're cycling through edges from IN vertices)

					//
					// The "normal" case of an external node connected to an internal node
					// CASE: OUT-IN
					if ((nodes.Contains(e.v1) && (!nodes.Contains(e.v2) && (newguy != e.v2)))
						|| (nodes.Contains(e.v2) && (!nodes.Contains(e.v1) && (newguy != e.v1))))
					{
						Vertex outVertex = null;
						Vertex inVertex = null;

						if (!nodes.Contains(e.v1))
						{
							// e.v1 is the OUT node
							outVertex = e.v1;
							inVertex = e.v2;
						}
						else
						{
							outVertex = e.v2;
							inVertex = e.v1;
						}

						//
						// Check to see if newguy is already connected to out node
						if (!newguy.ConnectedTo.Contains(outVertex))
						{
							// newguy is NOT already connected to this vertex
							inVertex.removeRefToEdge(e);
							newguy.addRefToEdge(e);

							if (e.v1 == inVertex)
							{
								e.v1 = newguy;
							}
							else
							{
								e.v2 = newguy;
							}
						}
						else
						{
							// newguy IS already connected to this vertex (edge needs to be deleted)
							inVertex.removeRefToEdge(e);
							outVertex.removeRefToEdge(e);
							edgesToDelete.Add(e);

							newguy.getEdgeWithVertex(outVertex).weight += e.weight;
						}

						didSomethingWithThisEdge = true;
					}

					if (didSomethingWithThisEdge == false)
					{
						throw new Exception("collapseNodes: Mishandled an edge case when collapsing nodes.");
					}
				}

				foreach(Edge e in edgesToDelete)
				{
					this.removeEdge(e.v1, e.v2);
					newguy.removeRefToEdge(e);
				}
			}

			//
			// Update connected component
			//
			ConnectedComponent cc = newguy.ConnectedComponent;
			
			cc.Vertices.Add(newguy.ID, newguy);
			foreach(Vertex v in nodes)
			{
				cc.Vertices.Remove(v.ID);
			}

			cc.createAdjMatrix(newguy.ConnectedComponent.Vertices.Count);
			cc.calculateVertexComponentInformation(this.pb);

			return newguy;
		}

		/// <summary>
		/// Performs a dfs on this graph.
		/// </summary>
		/// <param name="v">Vertex to dfs on.</param>
		/// <param name="vertices">ArrayList to add found vertices to.</param>
		private void dfs(Vertex v, ArrayList vertices) 
		{
			foreach(Vertex u in v.ConnectedTo) 
			{
				if (u.marked == false) 
				{
					vertices.Add(u);
					u.marked = true;
					dfs(u, vertices);
				}
			}
		}

		/// <summary>
		/// Adds a vertex to this graph (and created the vertex, too).
		/// OVERLOADED
		/// </summary>
		/// <param name="id">The id number of the new node.</param>
		/// <param name="str">The label of the new node.</param>
		/// <param name="isChemical">Whether the new node is a chemical or not.</param>
		public Vertex addVertex(int id, string str, bool isChemical) 
		{
			Vertex newVertex = new Vertex(id, str, isChemical);
			_vertices.Add(id, newVertex);
			return newVertex;
		}

		/// <summary>
		/// Adds a vertex to this graph by reference.
		/// OVERLOADED
		/// </summary>
		/// <param name="v1">The vertex to add.</param>
		public void addVertex(Vertex v1) 
		{
			_vertices.Add(v1.ID, v1);
		}

		/// <summary>
		/// Adds an edge to this graph.
		/// </summary>
		/// <param name="v1">The first vertex in this edge.</param>
		/// <param name="v2">The second vertex in this edge.</param>
		/// <param name="str">This edge's string type (not necessary).</param>
		public void addEdge(Vertex v1, Vertex v2, string str, double weight) 
		{
			foreach(Edge e in _edges) 
			{
				if (((e.v1 == v1) && (e.v2 == v2)) || ((e.v1 == v2) && (e.v2 == v1))) 
				{
					return; // already in there
				}
			}

			_edges.Add(new Edge(v1, v2, str, weight));
		}

		/// <summary>
		/// Add an edge to this graph (SEARCHES FOR VERTICES)
		/// </summary>
		/// <param name="id1">the ID of one of the vertices in the edge</param>
		/// <param name="id2">the ID of the other vertex in the edge</param>
		/// <param name="str">the type of the edge</param>
		public void addEdge(int id1, int id2, string str, double weight) 
		{
			Vertex v1;
			Vertex v2;

			v1 = (Vertex) _vertices[id1];
			v2 = (Vertex) _vertices[id2];

			if ((v1 != null) && (v2 != null) && (str != null)) 
			{
				foreach(Edge e in _edges) 
				{
					if (((e.v1 == v1) && (e.v2 == v2)) || ((e.v1 == v2) && (e.v2 == v1))) 
					{
						return; // already in there
					}
				}

				_edges.Add(new Edge(v1, v2, str, weight));
			} 
			else 
			{
				throw new ArgumentException("Either one of the IDs you passed isn't paired with a vertex, or the edge type was null.");
			}
		}

		/// <summary>
		/// Remove an edge containing the two specified vertices.  UpdateComponents() must be called outside of this function.
		/// </summary>
		/// <param name="v1">First vertex.</param>
		/// <param name="v2">Second vertex.</param>
		public void removeEdge(Vertex v1, Vertex v2)
		{
			Edge toDelete = null;

			foreach (Edge e in _edges)
			{
				if (((e.v1 == v1) && (e.v2 == v2)) ||
					((e.v2 == v1) && (e.v1 == v2)))
				{
					toDelete = e;
				}
			}

			if (toDelete != null)
			{
				_edges.Remove(toDelete);
				toDelete.v1.removeRefToEdge(toDelete);
				toDelete.v2.removeRefToEdge(toDelete);
			}

			//
			// DON'T ALWAYS NEED TO UPDATE COMPONENTS, so leave to calling function
			//
		}

		/// <summary>
		/// Remove an edge containing the two specified vertices.  UpdateComponents() must be called outside of this function.
		/// </summary>
		/// <param name="id1">First vertex's ID.</param>
		/// <param name="id2">Second vertex's ID.</param>
		public void removeEdge(int id1, int id2)
		{
			Edge toDelete = null;
			Vertex v1 = (Vertex)_vertices[id1];
			Vertex v2 = (Vertex)_vertices[id2];

			foreach (Edge e in _edges)
			{
				if (((e.v1 == v1) && (e.v2 == v2)) ||
					((e.v2 == v1) && (e.v1 == v2)))
				{
					toDelete = e;
				}
			}

			if (toDelete != null)
			{
				_edges.Remove(toDelete);
			}

			//
			// DON'T ALWAYS NEED TO UPDATE COMPONENTS, so leave to calling function
			//
		}

		/// <summary>
		/// Draws the Fruchterman-Reingold representation of the graph
		/// </summary>
		public Bitmap drawGraph() 
		{
			Pen whitePen = new Pen(Color.White, 1);
			Pen grayPen = new Pen(Color.Gray, 1);
			Pen blackPen = new Pen(Color.Black, 1);
			Pen bigBlackPen = new Pen(Color.Black, 2);
			Pen dgPen = new Pen(Color.DarkGray, 1);
			Pen pinkPen = new Pen(Color.Pink, 1);
			Pen orangePen = new Pen(Color.Orange, 2);
			Pen redPen = new Pen(Color.Red, 1);
			Brush whiteBrush = new SolidBrush(Color.White);
			Brush grayBrush = new SolidBrush(Color.Gray);
			Brush redBrush = new SolidBrush(Color.FromArgb(255, 51, 0));
			Brush greenBrush = new SolidBrush(Color.Green);
			Brush blackBrush = new SolidBrush(Color.Black);
			Brush orangeBrush = new SolidBrush(Color.Orange);
			Brush pinkBrush = new SolidBrush(Color.Pink);
			Brush darkGreenBrush = new SolidBrush(Color.DarkGreen);
			Brush siennaBrush = new SolidBrush(Color.FromArgb(0, 51, 153));

			// This is the lightest color that shows up as a solid line
			//Brush lightGBrush = new SolidBrush(Color.FromArgb(204,204,204));
			//Pen lightGPen = new Pen(lightGBrush, 1);
			// This is the darkest color to use for an edge
			//Brush darkGBrush = new SolidBrush(Color.FromArgb(50,50,50));
			//Pen darkGPen = new Pen(darkGBrush, 1);

			int w = this.mainDisplayWidth;
			int h = this.mainDisplayHeight;
			Bitmap pic = new Bitmap(w, h);
			Graphics pad = Graphics.FromImage(pic);

			pad.FillRectangle(whiteBrush, 0, 0, w, h);
			pad.DrawRectangle(grayPen, 0, 0, w - 1, h - 1);

			Vertex selectedVertex = this.SelectedVertex;

			double minEdgeWeight = this.minEdgeWeight;
			double maxEdgeWeight = this.maxEdgeWeight;
			
			//
			//
			// Draw edges (some will be overdrawn later)
			if (this.drawOtherEdges == true)
			{
				if (minEdgeWeight == maxEdgeWeight)
				{
					//
					// Draw every line with equal darkness
					foreach(Edge e in this.Edges) 
					{
						if ((e.v1 != selectedVertex) && (e.v2 != selectedVertex))
						{
							pad.DrawLine(dgPen, (int)e.v1.X, (int)e.v1.Y, (int)e.v2.X, (int)e.v2.Y);
						}
					}
				}
				else
				{
					//
					// Scale the color such that minWeight = 204; maxWeight = 50
					// This yields: norm. distance from bottom = (w - minW) / (maxW - minW)
					//				color to use = 204 - ((w - minW) / (maxW - minW)) * 154
					foreach(Edge e in this.Edges) 
					{
						if ((e.v1 != selectedVertex) && (e.v2 != selectedVertex))
						{
							double rgbVal = 204.0 - (((e.weight - minEdgeWeight) / (maxEdgeWeight - minEdgeWeight)) * 154.0);
							pad.DrawLine(new Pen(new SolidBrush(Color.FromArgb((int)rgbVal,(int)rgbVal,(int)rgbVal))),
								(int)e.v1.X, (int)e.v1.Y, (int)e.v2.X, (int)e.v2.Y);
						}
					}
				}
			}

			//
			//
			// Draw unconnected vertices and vertices in connected component
			foreach(DictionaryEntry de in this.Vertices) 
			{
				Vertex v = (Vertex)(de.Value);

				if ((v != selectedVertex)
					&& (!selectedVertex.ConnectedTo.Contains(v))
					&& (selectedVertex.ConnectedComponent.Vertices.ContainsKey(v.ID))) 
				{
					pad.FillEllipse(siennaBrush, (int)v.X - 4, (int)v.Y - 4, 8, 8);
				} 
				else if ((v != selectedVertex) 
					&& (!selectedVertex.ConnectedTo.Contains(v))) 
				{
					pad.FillEllipse(blackBrush, (int)v.X - 4, (int)v.Y - 4, 8, 8);
				}
			}

			//
			//
			// Draw edges connected to selected vertex
			if (this.drawConnectedEdges == true)
			{
				if (minEdgeWeight == maxEdgeWeight)
				{
					foreach(Edge e in selectedVertex.InEdges) 
					{
						pad.DrawLine(orangePen, (int)e.v1.X, (int)e.v1.Y, (int)e.v2.X, (int)e.v2.Y);
					}
				}
				else
				{
					foreach(Edge e in selectedVertex.InEdges) 
					{
						double rVal = (double)Color.Yellow.R - (((e.weight - minEdgeWeight) / 
							(maxEdgeWeight - minEdgeWeight)) * (double)(Color.Yellow.R - Color.DarkOrange.R));
						double gVal = (double)Color.Yellow.G - (((e.weight - minEdgeWeight) / 
							(maxEdgeWeight - minEdgeWeight)) * (double)(Color.Yellow.G - Color.DarkOrange.G));
						double bVal = (double)Color.Yellow.B - (((e.weight - minEdgeWeight) / 
							(maxEdgeWeight - minEdgeWeight)) * (double)(Color.Yellow.B - Color.DarkOrange.B));
					
						pad.DrawLine(new Pen(new SolidBrush(Color.FromArgb((int)rVal, (int)gVal, (int)bVal)), 2), 
							(int)e.v1.X, (int)e.v1.Y, (int)e.v2.X, (int)e.v2.Y);
					}
				}
			}

			//
			// 
			// Draw text labels
			foreach(DictionaryEntry de in this.Vertices)
			{
				Vertex v = (Vertex)de.Value;

				if (v != selectedVertex)
				{
					pad.DrawString(v.Label, new Font(FontFamily.GenericSansSerif, 7, FontStyle.Bold), 
						darkGreenBrush, new PointF((float)(v.X + 0.0), (float)(v.Y + 0.0)));
				}
			}

			//
			//
			// Draw vertices connected to selected vertex
			foreach(DictionaryEntry de in this.Vertices) 
			{
				Vertex v = (Vertex)(de.Value);

				if (selectedVertex.ConnectedTo.Contains(v)) 
				{
					pad.FillEllipse(redBrush, (int)v.X - 4, (int)v.Y - 4, 8, 8);
				}
			}

			//
			//
			// Now color selected vertex
			PointF[] points = new PointF[3];
			points[0] = new PointF((float)selectedVertex.X, (float)selectedVertex.Y - (float)(10.0f*Math.Sqrt(3)*(1.0f/3.0f)));
			points[1] = new PointF((float)selectedVertex.X - 7.0f, (float)selectedVertex.Y + (float)(5.0f*Math.Sqrt(3)));
			points[2] = new PointF((float)selectedVertex.X + 7.0f, (float)selectedVertex.Y + (float)(5.0f*Math.Sqrt(3)));
			pad.FillPolygon(redBrush, points);
			pad.DrawString(selectedVertex.Label, new Font(FontFamily.GenericSansSerif, 8, FontStyle.Bold), 
				darkGreenBrush, new PointF((float)(selectedVertex.X + 3.0), (float)(selectedVertex.Y - 9.0)));

			return pic;
		}

		/// <summary>
		/// Draws the Fruchterman-Reingold representation of the graph
		/// </summary>
		public Bitmap drawPcaGraph()
		{
			Pen whitePen = new Pen(Color.White, 1);
			Pen grayPen = new Pen(Color.Gray, 1);
			Pen blackPen = new Pen(Color.Black, 1);
			Pen bigBlackPen = new Pen(Color.Black, 2);
			Pen dgPen = new Pen(Color.DarkGray, 1);
			Pen pinkPen = new Pen(Color.Pink, 1);
			Pen orangePen = new Pen(Color.Orange, 2);
			Pen redPen = new Pen(Color.Red, 1);
			Brush whiteBrush = new SolidBrush(Color.White);
			Brush grayBrush = new SolidBrush(Color.Gray);
			Brush redBrush = new SolidBrush(Color.FromArgb(255, 51, 0));
			Brush greenBrush = new SolidBrush(Color.Green);
			Brush blackBrush = new SolidBrush(Color.Black);
			Brush orangeBrush = new SolidBrush(Color.Orange);
			Brush pinkBrush = new SolidBrush(Color.Pink);
			Brush darkGreenBrush = new SolidBrush(Color.DarkGreen);
			Brush siennaBrush = new SolidBrush(Color.FromArgb(0, 51, 153));

			// This is the lightest color that shows up as a solid line
			//Brush lightGBrush = new SolidBrush(Color.FromArgb(204,204,204));
			//Pen lightGPen = new Pen(lightGBrush, 1);
			// This is the darkest color to use for an edge
			//Brush darkGBrush = new SolidBrush(Color.FromArgb(50,50,50));
			//Pen darkGPen = new Pen(darkGBrush, 1);

			int w = this.mainDisplayWidth;
			int h = this.mainDisplayHeight;
			Bitmap pic = new Bitmap(w, h);
			Graphics pad = Graphics.FromImage(pic);

			pad.FillRectangle(whiteBrush, 0, 0, w, h);
			pad.DrawRectangle(grayPen, 0, 0, w - 1, h - 1);

			Vertex selectedVertex = this.SelectedVertex;

			double minEdgeWeight = this.minEdgeWeight;
			double maxEdgeWeight = this.maxEdgeWeight;
			
			//
			//
			// Draw edges (some will be overdrawn later)
			if (this.drawOtherEdges == true)
			{
				if (minEdgeWeight == maxEdgeWeight)
				{
					//
					// Draw every line with equal darkness
					foreach(Edge e in this.Edges) 
					{
						if ((e.v1 != selectedVertex) && (e.v2 != selectedVertex))
						{
							pad.DrawLine(dgPen, (int)e.v1.XPca, (int)e.v1.YPca, (int)e.v2.XPca, (int)e.v2.YPca);
						}
					}
				}
				else
				{
					//
					// Scale the color such that minWeight = 204; maxWeight = 50
					// This yields: norm. distance from bottom = (w - minW) / (maxW - minW)
					//				color to use = 204 - ((w - minW) / (maxW - minW)) * 154
					foreach(Edge e in this.Edges) 
					{
						if ((e.v1 != selectedVertex) && (e.v2 != selectedVertex))
						{
							double rgbVal = 204.0 - (((e.weight - minEdgeWeight) / (maxEdgeWeight - minEdgeWeight)) * 154.0);
							pad.DrawLine(new Pen(new SolidBrush(Color.FromArgb((int)rgbVal,(int)rgbVal,(int)rgbVal))),
								(int)e.v1.XPca, (int)e.v1.YPca, (int)e.v2.XPca, (int)e.v2.YPca);
						}
					}
				}
			}

			//
			//
			// Draw unconnected vertices and vertices in connected component
			foreach(DictionaryEntry de in this.Vertices) 
			{
				Vertex v = (Vertex)(de.Value);

				if ((v != selectedVertex)
					&& (!selectedVertex.ConnectedTo.Contains(v))
					&& (selectedVertex.ConnectedComponent.Vertices.ContainsKey(v.ID))) 
				{
					pad.FillEllipse(siennaBrush, (int)v.XPca - 4, (int)v.YPca - 4, 8, 8);
				} 
				else if ((v != selectedVertex) 
					&& (!selectedVertex.ConnectedTo.Contains(v))) 
				{
					pad.FillEllipse(blackBrush, (int)v.XPca - 4, (int)v.YPca - 4, 8, 8);
				}
			}

			//
			//
			// Draw edges connected to selected vertex
			if (this.drawConnectedEdges == true)
			{
				if (minEdgeWeight == maxEdgeWeight)
				{
					foreach(Edge e in selectedVertex.InEdges) 
					{
						pad.DrawLine(orangePen, (int)e.v1.XPca, (int)e.v1.YPca, (int)e.v2.XPca, (int)e.v2.YPca);
					}
				}
				else
				{
					foreach(Edge e in selectedVertex.InEdges) 
					{
						double rVal = (double)Color.Yellow.R - (((e.weight - minEdgeWeight) / 
							(maxEdgeWeight - minEdgeWeight)) * (double)(Color.Yellow.R - Color.DarkOrange.R));
						double gVal = (double)Color.Yellow.G - (((e.weight - minEdgeWeight) / 
							(maxEdgeWeight - minEdgeWeight)) * (double)(Color.Yellow.G - Color.DarkOrange.G));
						double bVal = (double)Color.Yellow.B - (((e.weight - minEdgeWeight) / 
							(maxEdgeWeight - minEdgeWeight)) * (double)(Color.Yellow.B - Color.DarkOrange.B));
					
						pad.DrawLine(new Pen(new SolidBrush(Color.FromArgb((int)rVal, (int)gVal, (int)bVal)), 2), 
							(int)e.v1.XPca, (int)e.v1.YPca, (int)e.v2.XPca, (int)e.v2.YPca);
					}
				}
			}

			//
			// 
			// Draw text labels
			foreach(DictionaryEntry de in this.Vertices)
			{
				Vertex v = (Vertex)de.Value;

				if (v != selectedVertex)
				{
					pad.DrawString(v.Label, new Font(FontFamily.GenericSansSerif, 7, FontStyle.Bold), 
						darkGreenBrush, new PointF((float)(v.XPca + 0.0), (float)(v.YPca + 0.0)));
				}
			}

			//
			//
			// Draw vertices connected to selected vertex
			foreach(DictionaryEntry de in this.Vertices) 
			{
				Vertex v = (Vertex)(de.Value);

				if (selectedVertex.ConnectedTo.Contains(v)) 
				{
					pad.FillEllipse(redBrush, (int)v.XPca - 4, (int)v.YPca - 4, 8, 8);
				}
			}

			//
			//
			// Now color selected vertex
			PointF[] points = new PointF[3];
			points[0] = new PointF((float)selectedVertex.XPca, (float)selectedVertex.YPca - (float)(10.0f*Math.Sqrt(3)*(1.0f/3.0f)));
			points[1] = new PointF((float)selectedVertex.XPca - 7.0f, (float)selectedVertex.YPca + (float)(5.0f*Math.Sqrt(3)));
			points[2] = new PointF((float)selectedVertex.XPca + 7.0f, (float)selectedVertex.YPca + (float)(5.0f*Math.Sqrt(3)));
			pad.FillPolygon(redBrush, points);
			pad.DrawString(selectedVertex.Label, new Font(FontFamily.GenericSansSerif, 8, FontStyle.Bold), 
				darkGreenBrush, new PointF((float)(selectedVertex.XPca + 3.0), (float)(selectedVertex.YPca - 9.0)));

			return pic;
		}

		/// <summary>
		/// Calculates graph coordinates in the Laplacian LLE graph for each vertex.
		/// </summary>
		public void calculatePcaGraphCoords()
		{
			double xScale = 1.0;
			double yScale = 1.0;

			double numVertices = Vertices.Count;

			int w = this.mainDisplayWidth;
			int h = this.mainDisplayHeight;

			//
			// Set original graph positions
			foreach (DictionaryEntry de in Vertices)
			{
				Vertex v = (Vertex)de.Value;

				v.XPca = v.getPcaValue(0);
				v.YPca = v.getPcaValue(1);
			}

			//
			// Scale graph
			//
			// Find x, y scales
			double minX = Double.MaxValue;
			double maxX = Double.MinValue;
			double minY = Double.MaxValue;
			double maxY = Double.MinValue;
			foreach(DictionaryEntry de in Vertices) 
			{
				Vertex v = (Vertex)(de.Value);

				if (v.XPca < minX) 
				{
					minX = v.XPca;
				}

				if (v.XPca > maxX) 
				{
					maxX = v.XPca;
				}

				if (v.YPca < minY) 
				{
					minY = v.YPca;
				}

				if (v.YPca > maxY) 
				{
					maxY = v.YPca;
				}
			}

			foreach(DictionaryEntry de in Vertices) 
			{
				Vertex v = (Vertex)(de.Value);

				v.XPca -= (maxX + minX) / 2.0;
				v.YPca -= (maxY + minY) / 2.0;
			}
				
			xScale = (((double) w)*.5 - 20) / (maxX - ((maxX + minX) / 2.0));
			yScale = (((double) h)*.5 - 20) / (maxY - ((maxY + minY) / 2.0));

			// Scale vertices
			foreach(DictionaryEntry de in Vertices) 
			{
				Vertex v = (Vertex)(de.Value);

				v.XPca = v.XPca * xScale + (((double) w)*.5 - 5);
				v.YPca = v.YPca * yScale + (((double) h)*.5 - 5);
			}
		}
	}
}
