using System;
using System.Collections;

namespace SpectralNet
{
	/// <summary>
	/// A node in a graph
	/// </summary>
	public class Vertex 
	{
		/// <summary>
		/// This node's identification number.  These numbers may be arbitrary, as long as they are unique over the graph.
		/// </summary>
		int _id;
		
		/// <summary>
		/// The string representing this node. I.e. "4 H11".
		/// </summary>
		string _label;

		/// <summary>
		/// The number of edges this vertex is connected to.  Unused?
		/// </summary>
		int _numRefs = 0;

		/// <summary>
		/// Distance (in hops) from this vertex to the rest of the graph.
		/// </summary>
		double _minDist, _maxDist, _avgDist;

		/// <summary>
		/// Coordinates for avg. path length graph.
		/// </summary>
		double _xAvgDist, _yAvgDist;

		/// <summary>
		/// Represents whether this node is internally regarded as a chemical.  Used in pseudo-modal projection w/ regex.
		/// </summary>
		bool _isChemical;

		/// <summary>
		/// Used in making hierarchical graphs, this represents whether this node is a peripheral node.
		/// </summary>
		bool _isPeripheral;

		/// <summary>
		/// Used to mark during searching.
		/// </summary>
		public bool marked;

		/// <summary>
		/// Used in running F-R algorithm.
		/// </summary>
		double _xdisp, _ydisp;

		/// <summary>
		/// Graph coordinates in main graph.
		/// </summary>
		double _xpos, _ypos;
		double _xposOrig, _yposOrig;

		/// <summary>
		/// Graph coordinates in Laplacian LLE graph.
		/// </summary>
		double _xLap, _yLap;
		double _xLapOrig, _yLapOrig;

		/// <summary>
		/// Graph coordinates in PCA graph.
		/// </summary>
		double _xPca, _yPca;
		double _xPcaOrig, _yPcaOrig;

		/// <summary>
		/// VRML coordinates, scaled to 4.0, of the vertex
		/// </summary>
		double _xVRML, _yVRML, _zVRML;

		/// <summary>
		/// This node's cluster coefficient, which measures the proportion of neighbor nodes that are connected to each other.
		/// </summary>
		double _clusterCoeff;

		/// <summary>
		/// Graph coordinates in the cluster coefficient graph.
		/// </summary>
		double _xCC, _yCC;

		/// <summary>
		/// List of edges this vertex is a member of.
		/// </summary>
		ArrayList _inEdges;

		/// <summary>
		/// Records any vertices collapsed into this vertex
		/// </summary>
		ArrayList _childVertices;

		/// <summary>
		/// Whether or not this vertex has been collapsed into a parent vertex.
		/// </summary>
		bool _isChild;

		/// <summary>
		/// If this is a child vertex, this is a list of the vertices this vertex used to be connected to.
		/// </summary>
		ArrayList _wasConnectedTo;

		/// <summary>
		/// Adjacency eigenvector coordinates.
		/// </summary>
		EigenInformation _adjEigen;

		/// <summary>
		/// Laplacian eigenvector coordinates.
		/// </summary>
		EigenInformation _lapEigen;

		/// <summary>
		/// Normalized Laplacian eigenvector coordinates.
		/// </summary>
		EigenInformation _normLapEigen;

		/// <summary>
		/// Whether this vertex has component information calculated (cc must be large enough).
		/// </summary>
		public bool hasComponentInfo;

		/// <summary>
		/// The connected component this vertex belongs to.
		/// </summary>
		ConnectedComponent _connComp;

		/// <summary>
		/// Only used for comparing two graphs -- this matches vertices from one graph to the other.
		/// </summary>
		public Vertex partnerVertex;

		/// <summary>
		/// Only used for comparing two graphs -- this is the distance from a node to its homologue
		/// </summary>
		public double parallelDistance;

		/// <summary>
		/// Gets the specified PCA coordinate.
		/// </summary>
		/// <param name="which">Which coordinate to return.</param>
		/// <returns>The PCA coordinate from the scores matrix.</returns>
		public double getPcaValue(int which)
		{
			return _connComp.Graph.pcaScores[_connComp.Vertices.IndexOfKey(_id), which];
		}

		/// <summary>
		/// Gets the specified adjacency eigenvector component.
		/// </summary>
		/// <param name="which">Which eigenvalue's eigenvector component to return.</param>
		/// <returns>The eigenvector component.</returns>
		public double getAdjacencyEigenvalue(int which)
		{
			return _connComp.adjEigenvectors[_connComp.Vertices.IndexOfKey(_id), which];
		}

		/// <summary>
		/// Gets the specified laplacian eigenvector component.
		/// </summary>
		/// <param name="which">Which eigenvalue's eigenvector component to return.</param>
		/// <returns>The eigenvector component.</returns>
		public double getLaplacianEigenvalue(int which)
		{
			return _connComp.lapEigenvectors[_connComp.Vertices.IndexOfKey(_id), which];
		}

		/// <summary>
		/// Gets the specified normalized laplacian eigenvector component.
		/// </summary>
		/// <param name="which">Which eigenvalue's eigenvector component to return.</param>
		/// <returns>The eigenvector component.</returns>
		public double getNormalizedLaplacianEigenvalue(int which)
		{
			return _connComp.normLapEigenvectors[_connComp.Vertices.IndexOfKey(_id), which];
		}

		public NodeDoubleVector getLaplacianAsNodeDoubleVector()
		{
			NodeDoubleVector result = new NodeDoubleVector(4);
			for(int i = 1; i < 4; i++)
			{
				result.Values[i] = this.getLaplacianEigenvalue(this.ConnectedComponent.Vertices.Count - i);
			}
			return result;

			/*
			NodeDoubleVector result = new NodeDoubleVector(this.ConnectedComponent.NumVertices);
			for(int i = 0; i < this.ConnectedComponent.NumVertices; i++)
			{
				result.Values[i] = this.getLaplacianEigenvalue(i);
			}
			return result;
			*/
		}

		/// <summary>
		/// If this vertex is a Child Vertex, returns nodes this vertex used to be
		/// connected to.  If this vertex is not a Child Vertex, throws an error.
		/// </summary>
		public ArrayList WasConnectedTo
		{
			get
			{
				if (_isChild == false)
				{
					throw new Exception("Trying to get WasConnectedTo nodes for a vertex that is not a child.");
				}
				else
				{
					return _wasConnectedTo;
				}
			}
			set
			{
				if (_isChild == false)
				{
					throw new Exception("Trying to set WasConnectedTo nodes for a vertex that is not a child.");
				}
				else
				{
					_wasConnectedTo = value;
				}
			}
		}

		public double Degree
		{
			get
			{
				double result = 0;

				foreach(Edge e in this.InEdges)
				{
					result += e.weight;
				}

				return result;
			}
		}

		/// <summary>
		/// Property wrapper for _isChild
		/// </summary>
		public bool IsChild
		{
			get { return _isChild; }
			set { _isChild = value; }
		}

		/// <summary>
		/// Property wrapper for _childVertices
		/// </summary>
		public ArrayList ChildVertices
		{
			get { return _childVertices; }
			set { _childVertices = value; }
		}

		/// <summary>
		/// Property wrapper for _xLap
		/// </summary>
		public double XLap
		{
			get { return _xLap; }
			set { _xLap = value; }
		}

		/// <summary>
		/// Property wrapper for _yLap
		/// </summary>
		public double YLap
		{
			get { return _yLap; }
			set { _yLap = value; }
		}

		/// <summary>
		/// Property wrapper for _xLapOrig
		/// </summary>
		public double XLapOrig
		{
			get { return _xLapOrig; }
			set { _xLapOrig = value; }
		}

		/// <summary>
		/// Property wrapper for _yLapOrig
		/// </summary>
		public double YLapOrig
		{
			get { return _yLapOrig; }
			set { _yLapOrig = value; }
		}

		/// <summary>
		/// Property wrapper for _xPca
		/// </summary>
		public double XPca
		{
			get { return _xPca; }
			set { _xPca = value; }
		}

		/// <summary>
		/// Property wrapper for _yPca
		/// </summary>
		public double YPca
		{
			get { return _yPca; }
			set { _yPca = value; }
		}

		/// <summary>
		/// Property wrapper for _xPcaOrig
		/// </summary>
		public double XPcaOrig
		{
			get { return _xPcaOrig; }
			set { _xPcaOrig = value; }
		}

		/// <summary>
		/// Property wrapper for _yPcaOrig
		/// </summary>
		public double YPcaOrig
		{
			get { return _yPcaOrig; }
			set { _yPcaOrig = value; }
		}

		/// <summary>
		/// Property wrapper for _xVRML
		/// </summary>
		public double XVRML
		{
			get { return _xVRML; }
			set { _xVRML = value; }
		}

		/// <summary>
		/// Property wrapper for _yVRML
		/// </summary>
		public double YVRML
		{
			get { return _yVRML; }
			set { _yVRML = value; }
		}

		/// <summary>
		/// Property wrapper for _zVRML
		/// </summary>
		public double ZVRML
		{
			get { return _zVRML; }
			set { _zVRML = value; }
		}

		/// <summary>
		/// Property wrapper for _isPeripheral
		/// </summary>
		public bool IsPeripheral
		{
			get { return _isPeripheral; }
			set { _isPeripheral = value; }
		}

		/// <summary>
		/// Property wrapper for _minDist
		/// </summary>
		public double MinDist 
		{
			get { return _minDist; }
			set { _minDist = value; }
		}

		/// <summary>
		/// Property wrapper for _maxDist
		/// </summary>
		public double MaxDist 
		{
			get { return _maxDist; }
			set { _maxDist = value; }
		}

		/// <summary>
		/// Property wrapper for _avgDist
		/// </summary>
		public double AvgDist 
		{
			get { return _avgDist; }
			set { _avgDist = value; }
		}

		/// <summary>
		/// Property wrapper for _clusterCoeff
		/// </summary>
		public double ClusterCoeff 
		{
			get { return _clusterCoeff; }
			set { _clusterCoeff = value; }
		}

		/// <summary>
		/// Property wrapper for _connComp
		/// </summary>
		public ConnectedComponent ConnectedComponent 
		{
			get { return _connComp; }
			set { _connComp = value; }
		}

		/// <summary>
		/// Property wrapper for _adjEigen
		/// </summary>
		public EigenInformation AdjacencyEigenInformation 
		{
			get { return _adjEigen; }
			set { _adjEigen = value; }
		}

		/// <summary>
		/// Property wrapper for _lapEigen
		/// </summary>
		public EigenInformation LaplacianEigenInformation 
		{
			get { return _lapEigen; }
			set { _lapEigen = value; }
		}

		/// <summary>
		/// Property wrapper for _normLapEigen
		/// </summary>
		public EigenInformation NormalizedLaplacianEigenInformation 
		{
			get { return _normLapEigen; }
			set { _normLapEigen = value; }
		}

		/// <summary>
		/// Property wrapper for _isChemical
		/// </summary>
		public bool isChemical 
		{
			get { return _isChemical; }
			set { _isChemical = value; }
		}

		/// <summary>
		/// Property wrapper for _label
		/// </summary>
		public string Label 
		{
			get { return _label; }
			set { _label = value; }
		}

		/// <summary>
		/// Property wrapper for _id
		/// </summary>
		public int ID 
		{
			get { return _id; }
			set { _id = value; }
		}

		/// <summary>
		/// Property wrapper for _xpos
		/// </summary>
		public double X 
		{
			get { return _xpos; }
			set { _xpos = value; }
		}

		/// <summary>
		/// Property wrapper for _ypos
		/// </summary>
		public double Y 
		{
			get { return _ypos; }
			set { _ypos = value; }
		}

		/// <summary>
		/// Property wrapper for _xposOrig
		/// </summary>
		public double XOrig
		{
			get { return _xposOrig; }
			set { _xposOrig = value; }
		}

		/// <summary>
		/// Property wrapper for _yposOrig
		/// </summary>
		public double YOrig
		{
			get { return _yposOrig; }
			set { _yposOrig = value; }
		}

		/// <summary>
		/// Property wrapper for _xCC
		/// </summary>
		public double xCC 
		{
			get { return _xCC; }
			set { _xCC = value; }
		}

		/// <summary>
		/// Property wrapper for _yCC
		/// </summary>
		public double yCC 
		{
			get { return _yCC; }
			set { _yCC = value; }
		}

		/// <summary>
		/// Property wrapper for _xAvgDist
		/// </summary>
		public double xAverageDistance
		{
			get { return _xAvgDist; }
			set { _xAvgDist = value; }
		}

		/// <summary>
		/// Property wrapper for _yAvgDist
		/// </summary>
		public double yAverageDistance
		{
			get { return _yAvgDist; }
			set { _yAvgDist = value; }
		}

		/// <summary>
		/// Property wrapper for _xdisp
		/// </summary>
		public double dispX 
		{
			get { return _xdisp; }
			set { _xdisp = value; }
		}

		/// <summary>
		/// Property wrapper for _ydisp
		/// </summary>
		public double dispY 
		{
			get { return _ydisp; }
			set { _ydisp = value; }
		}

		/// <summary>
		/// Property wrapper for _inEdges
		/// </summary>
		public ArrayList InEdges 
		{
			get { return _inEdges; }
		}

		/// <summary>
		/// Vertex constructor
		/// </summary>
		/// <param name="id">This vertex's ID number.</param>
		/// <param name="n">This vertex's label.</param>
		/// <param name="isChemical">Whether this vertex is a chemical.</param>
		public Vertex(int id, string n, bool isChemical) 
		{
			Int64 numTicks = System.DateTime.Now.Ticks % 1000000;
			Random r = new Random(id + (int)numTicks);

			_id = id;
			_label = n;
			_isChemical = isChemical;
			_inEdges = new ArrayList();
			_xpos = r.NextDouble() - 0.5;
			_ypos = r.NextDouble() - 0.5;
			_xdisp = 0.0;
			_ydisp = 0.0;

			partnerVertex = null;
			parallelDistance = 0;

			_childVertices = new ArrayList();

			_isChild = false;

			hasComponentInfo = false;

			marked = false;

			r = null;
		}

		/// <summary>
		/// Adds a reference to an edge.
		/// </summary>
		/// <param name="ed">The edge to which to add a reference.</param>
		public void addRefToEdge(Edge ed) 
		{
			if (_inEdges.Contains(ed) == false) 
			{
				_inEdges.Add(ed);
				_numRefs++;
			}
		}

		/// <summary>
		/// Removes a reference to an edge.
		/// </summary>
		/// <param name="ed">The edge to which to remove a reference.</param>
		public void removeRefToEdge(Edge ed) 
		{
			if (_inEdges.Contains(ed) == true) 
			{
				_inEdges.Remove(ed);
				_numRefs--;
			}
		}

		/// <summary>
		/// Returns a list of vertices this vertex is connected to.
		/// </summary>
		public ArrayList ConnectedTo 
		{
			get 
			{
				ArrayList results = new ArrayList();
				
				foreach(Edge e in this._inEdges) 
				{
					if ((e.v1.Label == this.Label) && (e.v2.Label != this.Label)) 
					{
						results.Add(e.v2);
					} 
					else if ((e.v2.Label == this.Label) && (e.v1.Label != this.Label)) 
					{
						results.Add(e.v1);
					}
					else if ((e.v1.Label == this.Label) && (e.v2.Label == this.Label)) 
					{
						results.Add(e.v1);
					}
					else
					{
						// Neither vertex matches...this is an error
						throw new Exception("Vertex " + this.Label + " has an edge in _inEdges that it doesn't belong to.");
					}
				}

				return results;
			}
		}

		/// <summary>
		/// Returns the edge that this vertex is connected to containing the vertex v.  Throws an exception
		/// if the target vertex isn't found.
		/// </summary>
		/// <param name="v">Vertex to find connection to</param>
		/// <returns>The edge containing vertex v</returns>
		public Edge getEdgeWithVertex(Vertex v)
		{
			if (v.Label == this.Label)
			{
				foreach(Edge e in this._inEdges)
				{
					if ((e.v1 == v) && (e.v2 == v))
					{
						return e;
					}
				}
			}
			else
			{
				foreach(Edge e in this._inEdges)
				{
					if ((e.v1 == v) || (e.v2 == v))
					{
						return e;
					}
				}
			}

			throw new Exception("Cannot find vertex " + v.Label + " connected to vertex " + this.Label);
		}
	}
	
}
